(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)


structure SymbolTable (* :> SYMBOL_TABLE *) =
  struct
    type index = string
    type dimensions = index list
    datatype traversal = Arbitrary | Alphabetical
    type 'a t = 'a TernaryTree.t

    fun symbol name =
          name

    fun name theName =
          theName
    
    (* The rest implements symbol tables as
       ternary string trees *)
    fun empty () =
          TernaryTree.empty ()

    fun insert name binding aTable =
          TernaryTree.insert name binding aTable

    fun delete name aTable =
          TernaryTree.delete name aTable

    fun lookup aTable name =
          TernaryTree.lookup aTable name

    val tabulate = TernaryTree.tabulate
    val create = TernaryTree.create
    val sub = TernaryTree.sub
    val dimensions = TernaryTree.dimensions
    val foldli = TernaryTree.foldli
    val foldl = TernaryTree.foldl
    val mapi = TernaryTree.mapi
    val map = TernaryTree.map
    val append = TernaryTree.append
    val lookupKeyedPrefix = TernaryTree.lookupKeyedPrefix
  end

(*
functor FTestSymbolTable(s : SYMBOL_TABLE) =
  struct
    fun testIt () =
          let
            val a = s.symbol "test"
            val b = s.symbol "b"
            val t = s.empty
            val t1 = s.insert t a "this is the value of symbol a"
            val t2 = s.insert t1 b "this is the binding of symbol b"
          in
            (s.lookup t a) = NONE andalso
            (s.lookup t1 a) = SOME("this is the value of symbol a") andalso
            (s.lookup t1 b) = NONE andalso
            (s.lookup t2 b) = SOME("this is the binding of symbol b") andalso
            (s.lookup t2 a) = SOME("this is the value of symbol a")
          end
  end

structure TestSymbolTable = FTestSymbolTable(SymbolTable) *)

(* val _ = if (TestSymbolTable.testIt ()) then print "OK!" else (); *)