(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiAbs =
  struct
    type label = string
    type tyname = string
    type cname = string
    type tyvarname = string
    type metavarname = string
    type vid = string

    (* Restrictions for type variables *)
    datatype restrtypat =
      RstrTyPat of {args : {name : tyvarname, pat : restrtypat} list,
                    name : tyname, tags : cname list}

    datatype restr =
      RstrNone
    | RstrExclude of {labs : label list}
    | RstrTyPats of {pats : restrtypat list}
    
    (* Type system using restrictions *)
    datatype 'a ty =
      TyCon of {args : {name : tyvarname, ty : 'a ty} list,
                name : tyname, tags : cname list, ann : 'a}
    | TyFun of {from : 'a ty, to : 'a ty, ann : 'a}
    | TyField of {present : 'a ty, ty : 'a ty, ann : 'a}
    | TyExtRecord of {fields : {lab : label, ty : 'a ty, ann : 'a} list,
                      ext : 'a ty, ann : 'a}
    | TyEmptyRecord of {ann : 'a}
    | TyVar of {name : tyvarname, restr : restr, ann : 'a} 
    | TyMeta of {name : metavarname, restr : restr,
                 bound : 'a ty option ref, ann : 'a}
    
    type 'a tyscheme =
      {params : {name : tyvarname, restr : restr} list,
       ty : 'a ty, ann : 'a}

    (* Utility type for type inference *)
    type 'a orderrow =
      {lab : label, ty : unit ty, ann : 'a} list

    (* The base constants (These are really part of the supported basis) *)
    datatype scon =
      SCReal of real
    | SCInt of int
    | SCWord of word
    | SCString of string
    | SCChar of char

    (* The expression langauge *)
    (* Patterns *)
    datatype 'a pat =
      PatVar of {name : vid, ann : 'a}
    | PatTyped of {pat : 'a pat, ty : unit ty, ann : 'a}
    | PatLayered of {vid : vid, pat : 'a pat, ann : 'a}
    | PatCon of {tag : cname, tyScheme : unit tyscheme, ann : 'a}
    | PatParamCon of {tag : cname, tyScheme : unit tyscheme,
                      pat : 'a pat, ann : 'a}
    | PatEx of {tag : cname, ann : 'a}
    | PatParamEx of {tag : cname, pat : 'a pat, ty : unit ty, ann : 'a}
    | PatSCon of {scon : scon, ann : 'a}
    | PatWild of {ann : 'a}
    | PatEmptyRecord of {ann : 'a}
    | PatExtRecord of {fields : {lab : label, pat : 'a pat, ann : 'a} list,
                       ext : 'a pat, ann : 'a}
    | PatOptField of {present : 'a pat, pat : 'a pat, ann : 'a}

    (* Expressions *)
    datatype 'a exp =
      ExpFun of {param : vid, exp : 'a exp, ann : 'a}
    | ExpApp of {funExp : 'a exp, argExp : 'a exp, ann : 'a}
    | ExpVar of {name : vid, ann : 'a}
    | ExpTyped of {exp : 'a exp, ty : unit ty, ann : 'a}
    | ExpCon of {tag : cname, tyScheme : unit tyscheme, ann : 'a}
    | ExpEx of {tag : cname, ty : unit ty, ann : 'a}
    | ExpSCon of {scon : scon, ann : 'a}
    | ExpEmptyRecord of {ann : 'a}
    | ExpExtRecord of {fields : {lab : label, exp : 'a exp, ann : 'a} list,
                       ext : 'a exp, ann : 'a}
    | ExpOptField of {present : 'a exp, exp : 'a exp, ann : 'a}
    | ExpLet of {binds : {pat : 'a pat, exp : 'a exp, ann : 'a} list,
                 exp : 'a exp, ann : 'a}
    | ExpLetRec of {binds : {pat : 'a pat, exp : 'a exp, ann : 'a} list,
                    exp : 'a exp, ann : 'a}
    | ExpLetEx of {binds : {tag : cname, ty : unit ty, ann : 'a} list,
                   exp : 'a exp, ann : 'a}
    | ExpHandle of {exp : 'a exp,
                    matches : {pat : 'a pat, exp : 'a exp, ann : 'a} list,
                    ann : 'a}
    | ExpRaise of {exp : 'a exp, ann : 'a}
    | ExpCase of {exp : 'a exp,
                  matches : {pat : 'a pat, exp : 'a exp, ann : 'a} list,
                  ann : 'a}
    | ExpFieldCase of {exp : 'a exp, tyvar : tyvarname,
                       absExp : 'a exp,
                       prePat : 'a pat, preExp : 'a exp,
                       tyOut : unit ty, ann : 'a}
  end
