(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure TestKsiInterpret (* :> TESTER *) =
  struct
    structure ST = KsiInterpret.ST

    exception AcceptMoreThanOnce
    exception AcceptWrongValue
    exception BuiltinError

    val dummyAnn = {pos = {fileStartLoc = {line = 1, pos = 1},
                           fileEndLoc = {line = 1, pos = 1},
                           fileName = ""},
                    ty = KsiAbs.TyEmptyRecord {ann = ()}}

    (* \xi-Calculus abstract syntax for a dummy type scheme *)
    val dummyTyScheme = {params = nil,
                         ty = KsiAbs.TyEmptyRecord {ann = ()},
                         ann = ()}

    (* utility functions for building simple
       operator calls in \xi-Calculus abstract syntax. *)
    fun makePresentField exp =
          KsiAbs.ExpOptField
            {present = KsiAbs.ExpCon {tag = "present",
                                      tyScheme = dummyTyScheme,
                                      ann = dummyAnn},
             exp = exp, ann = dummyAnn}

    fun makeKsiExpOpExp exp1 opName exp2 =
          KsiAbs.ExpApp
            {funExp = KsiAbs.ExpVar
                        {name = opName,
                         ann = dummyAnn},
             argExp = KsiAbs.ExpExtRecord
                        {fields = [{lab = "1",
                                    exp = makePresentField exp1,
                                    ann = dummyAnn},
                                   {lab = "2",
                                    exp = makePresentField exp2,
                                    ann = dummyAnn}
                                  ],
                         ext = KsiAbs.ExpEmptyRecord {ann = dummyAnn},
                         ann = dummyAnn},
             ann = dummyAnn}

    fun makeKsiVarOpInt varName opName i =
          makeKsiExpOpExp
            (KsiAbs.ExpVar
               {name = varName, ann = dummyAnn})
            opName
            (KsiAbs.ExpSCon
               {scon = KsiAbs.SCInt i, ann = dummyAnn})

    fun makeKsiVarOpVar varName1 opName varName2 =
          makeKsiExpOpExp
            (KsiAbs.ExpVar
               {name = varName1, ann = dummyAnn})
            opName
            (KsiAbs.ExpVar
               {name = varName2, ann = dummyAnn})

    fun makeKsiExpConsExp exp1 tagName exp2 =
          KsiAbs.ExpApp
            {funExp = KsiAbs.ExpCon
                        {tag = tagName,
                         tyScheme = dummyTyScheme,
                         ann = dummyAnn},
             argExp = KsiAbs.ExpExtRecord
                        {fields = [{lab = "1",
                                    exp = makePresentField exp1,
                                    ann = dummyAnn},
                                   {lab = "2",
                                    exp = makePresentField exp2,
                                    ann = dummyAnn}
                                  ],
                         ext = KsiAbs.ExpEmptyRecord {ann = dummyAnn},
                         ann = dummyAnn},
             ann = dummyAnn}

    fun makeKsiVarConsVar varName1 tagName varName2 =
          makeKsiExpConsExp
            (KsiAbs.ExpVar
               {name = varName1, ann = dummyAnn})
            tagName
            (KsiAbs.ExpVar
               {name = varName2, ann = dummyAnn})


    fun makeKsiIntConsExp i tagName exp =
          makeKsiExpConsExp
            (KsiAbs.ExpSCon
               {scon = KsiAbs.SCInt i, ann = dummyAnn})
            tagName exp
            
    (* Value utility function *)
    fun decons value =
          (case value of
             (KsiVals.Constructor
                  {name = "::",
                   value = SOME (KsiVals.Record fields)}) =>
                     (case fields of
                        [{lab = "1", value = value1},
                         {lab = "2", value = value2}] =>
                           (value1, value2)
                      | [{lab = "2", value = value2},
                         {lab = "1", value = value1}] =>
                           (value1, value2)
                      | _ =>
                           (print ("KsiBuiltinTypeError: " ^ "decons (1)");
                            raise BuiltinError)
             )
           | _ =>
               (print ("KsiBuiltinTypeError: " ^ "decons (2)");
                raise BuiltinError)
          )
    
    (* In test1 we test the equivalent of this CeXL-program,
       where we assume the function accept5 to be predeclared.
       accept5 should be called exactly once with the integer argument 5. *)
    (*
       let
         fun testStop cnt =
               if cnt < 5 then
                 continue cnt
               else
                 accept5 cnt

         and continue cnt =
               testStop (cnt + 1)
       in
         testStop 0
       end
    *)
    
    (* This is the \xi-Calculus abstract syntax for the above program: *)
    val test1KsiExp =
          KsiAbs.ExpLetRec
            {binds = [{pat = KsiAbs.PatVar
                               {name = "testStop", ann = dummyAnn},
                       exp = KsiAbs.ExpFun
                               {param = "cnt",
                                exp = KsiAbs.ExpCase
                                        {exp = makeKsiVarOpInt
                                                 "cnt" "<" 5,
                                         matches =
                                           [{pat = KsiAbs.PatCon
                                                     {tag = "true",
                                                      tyScheme = dummyTyScheme,
                                                      ann = dummyAnn},
                                             exp = KsiAbs.ExpApp
                                                     {funExp = KsiAbs.ExpVar
                                                                 {name = "continue",
                                                                  ann = dummyAnn},
                                                      argExp = KsiAbs.ExpVar
                                                                 {name = "cnt",
                                                                  ann = dummyAnn},
                                                      ann = dummyAnn},
                                             ann = dummyAnn},
                                            {pat = KsiAbs.PatCon
                                                     {tag = "false",
                                                      tyScheme = dummyTyScheme,
                                                      ann = dummyAnn},
                                             exp = KsiAbs.ExpApp
                                                     {funExp = KsiAbs.ExpVar
                                                                 {name = "accept5",
                                                                  ann = dummyAnn},
                                                      argExp = KsiAbs.ExpVar
                                                                 {name = "cnt",
                                                                  ann = dummyAnn},
                                                      ann = dummyAnn},
                                             ann = dummyAnn}
                                           ],
                                         ann = dummyAnn},
                                ann = dummyAnn},
                       ann = dummyAnn},
                      {pat = KsiAbs.PatVar
                               {name = "continue", ann = dummyAnn},
                       exp = KsiAbs.ExpFun
                               {param = "cnt",
                                exp = KsiAbs.ExpApp
                                        {funExp = KsiAbs.ExpVar
                                                    {name = "testStop",
                                                     ann = dummyAnn},
                                         argExp = makeKsiVarOpInt
                                                    "cnt" "+" 1,
                                         ann = dummyAnn},
                                ann = dummyAnn},
                       ann = dummyAnn}],
             exp = KsiAbs.ExpApp
                     {funExp = KsiAbs.ExpVar
                                 {name = "testStop",
                                  ann = dummyAnn},
                      argExp = KsiAbs.ExpSCon
                                 {scon = KsiAbs.SCInt 0,
                                  ann = dummyAnn},
                      ann = dummyAnn},
             ann = dummyAnn}

    (* In test2 we test the equivalent of this CeXL-program,
       where we assume the function acceptList321 to be predeclared.
       acceptList321 should be called exactly once with the argument [3, 2, 1]. *)
    (*
       let
         fun rev l acc =
               case l of
                 (x::xs) =>
                   rev xs (x::acc)
               | nil =>
                   acc
       in
         acceptList321 (rev [1, 2, 3] nil)
       end
    *)
    
    (* This is the \xi-Calculus abstract syntax for the above program: *)
    val test2KsiExp =
          KsiAbs.ExpLetRec
            {binds =
               [{pat = KsiAbs.PatVar
                         {name = "rev", ann = dummyAnn},
                 exp = KsiAbs.ExpFun
                         {param = "l",
                          exp = KsiAbs.ExpFun
                                  {param = "acc",
                                   exp = KsiAbs.ExpCase
                                           {exp = KsiAbs.ExpVar
                                                    {name = "l", ann = dummyAnn},
                                            matches =
                                              [{pat = KsiAbs.PatParamCon
                                                        {tag = "::",
                                                         pat = KsiAbs.PatExtRecord
                                                                 {fields =
                                                                    [{lab = "1",
                                                                      pat = KsiAbs.PatOptField
                                                                              {present =
                                                                                 KsiAbs.PatCon
                                                                                   {tag = "present",
                                                                                    tyScheme = dummyTyScheme,
                                                                                    ann = dummyAnn},
                                                                               pat = KsiAbs.PatVar
                                                                                       {name = "x",
                                                                                        ann = dummyAnn},
                                                                               ann = dummyAnn},
                                                                      ann = dummyAnn},
                                                                     {lab = "2",
                                                                      pat = KsiAbs.PatOptField
                                                                              {present =
                                                                                 KsiAbs.PatCon
                                                                                   {tag = "present",
                                                                                    tyScheme = dummyTyScheme,
                                                                                    ann = dummyAnn},
                                                                               pat = KsiAbs.PatVar
                                                                                       {name = "xs",
                                                                                        ann = dummyAnn},
                                                                               ann = dummyAnn},
                                                                      ann = dummyAnn}],
                                                                  ext = KsiAbs.PatEmptyRecord
                                                                          {ann = dummyAnn},
                                                                  ann = dummyAnn},
                                                         tyScheme = dummyTyScheme,
                                                         ann = dummyAnn},
                                                exp = KsiAbs.ExpApp
                                                        {funExp = KsiAbs.ExpApp
                                                                    {funExp =
                                                                       KsiAbs.ExpVar
                                                                         {name = "rev",
                                                                          ann = dummyAnn},
                                                                     argExp =
                                                                       KsiAbs.ExpVar
                                                                         {name = "xs",
                                                                          ann = dummyAnn},
                                                                     ann = dummyAnn},
                                                         argExp = makeKsiVarConsVar
                                                                    "x" "::" "acc",
                                                         ann = dummyAnn},
                                                ann = dummyAnn},
                                               {pat = KsiAbs.PatCon
                                                        {tag = "nil",
                                                         tyScheme = dummyTyScheme,
                                                         ann = dummyAnn},
                                                exp = KsiAbs.ExpVar
                                                        {name = "acc",
                                                         ann = dummyAnn},
                                                ann = dummyAnn}
                                              ],
                                            ann = dummyAnn},
                                   ann = dummyAnn},
                          ann = dummyAnn},
                 ann = dummyAnn}],
             exp = KsiAbs.ExpApp
                     {funExp = KsiAbs.ExpVar
                                 {name = "acceptList321",
                                  ann = dummyAnn},
                      argExp = KsiAbs.ExpApp
                                 {funExp = KsiAbs.ExpApp
                                             {funExp =
                                                KsiAbs.ExpVar
                                                  {name = "rev",
                                                   ann = dummyAnn},
                                              argExp =
                                                makeKsiIntConsExp
                                                  1 "::"
                                                  (makeKsiIntConsExp
                                                     2 "::"
                                                     (makeKsiIntConsExp
                                                        3 "::"
                                                        (KsiAbs.ExpCon
                                                           {tag = "nil",
                                                            tyScheme = dummyTyScheme,
                                                            ann = dummyAnn})
                                                     )
                                                  ),
                                              ann = dummyAnn},
                                  argExp = KsiAbs.ExpCon
                                             {tag = "nil",
                                              tyScheme = dummyTyScheme,
                                              ann = dummyAnn},
                                  ann = dummyAnn},
                      ann = dummyAnn},
             ann = dummyAnn}
    
    (* Create the "built-in" environment and the test facilities *)
    val acceptCalled = ref false

    fun acceptInt i (KsiVals.Const (KsiAbs.SCInt value)) =
          if !acceptCalled then
            (* Out test-module treats a raised exception as an error *)
            raise AcceptMoreThanOnce
          else
            if i = value then
              (acceptCalled := true;
               KsiVals.Record nil)
            else
              (* Out test-module treats a raised exception as an error *)
              raise AcceptWrongValue
      | acceptInt i _ =
          raise BuiltinError

    fun acceptList321 value1 =
          if !acceptCalled then
            (* Out test-module treats a raised exception as an error *)
            (print "acceptList321 called more than once!";
             raise AcceptMoreThanOnce)
          else
            (case decons value1 of
               (KsiVals.Const (KsiAbs.SCInt 3),
                value2) =>
                  (case decons value2 of
                    (KsiVals.Const (KsiAbs.SCInt 2),
                     value3) =>
                       (case decons value3 of
                          (KsiVals.Const (KsiAbs.SCInt 1),
                           KsiVals.Constructor
                             {name = "nil", value = NONE}) =>
                                (acceptCalled := true;
                                 KsiVals.Record nil)
                        | _ =>
                            (* Out test-module treats a raised exception as an error *)
                            (print "acceptList321 not cons 1";
                             raise AcceptWrongValue)
                       )
                  | _ =>
                      (* Out test-module treats a raised exception as an error *)
                      (print "acceptList321 not cons 2";
                       raise AcceptWrongValue)
                 )
             | _ =>
                 (* Out test-module treats a raised exception as an error *)
                 (print "acceptList321 not cons 3";
                  raise AcceptWrongValue)
            )

    (* Build a symbol table of the environment of builtin functions *)
    fun bindFun env name funImpl =
          ST.insert (ST.symbol name) (KsiVals.Fun funImpl) env

    (* Our initial environment must have these "built-in" functions:
       < , + , accept5, acceptList321 *)
    fun makeInitEnv () =
          let
            val initEnv = KsiBuiltinDynamic.declareDynValEnv ()
            val initEnv = bindFun initEnv "accept5" (acceptInt 5)
            val initEnv = bindFun initEnv "acceptList321" acceptList321
          in
            initEnv
          end
    
    fun test1 () =
          (acceptCalled := false;
           KsiInterpret.evalExp (makeInitEnv ()) test1KsiExp;
           !acceptCalled)

    fun test2 () =
          (acceptCalled := false;
           KsiInterpret.evalExp (makeInitEnv ()) test2KsiExp;
           !acceptCalled)

    fun test () =
          TestUtil.doTest
            [("test1", test1),
             ("test2", test2)
            ]
  end