(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* The error reporting module for type-errors in
   CeXL / xi-Calculus-based input programs.
   An improved version of this module is used
   in the available binary file included with
   this implementation *)
structure KsiReportTypeErrors =
  struct
    exception InternalError

    (* Imperative variables *)
    val errorReportedRef = ref 0

    fun reset () =
          errorReportedRef := 0

    fun errorReported () =
          (!errorReportedRef) > 0

    fun reportError (_ : Word32.word) (_ : Word32.word) str =
          (errorReportedRef := (!errorReportedRef) + 1;
           print ("#" ^ (Int.toString (!errorReportedRef)) ^ ": " ^
                  str ^ "\n"))

    fun reportErrorStrs m code strs =
          reportError m code (String.concat strs)

    fun reportErrorFurther str =
          (if (!errorReportedRef) <= 0 then
             errorReportedRef := 1
           else
             ();
           print (str ^ "\n"))

    val reportErrorFurtherStrs =
          reportErrorFurther o String.concat

    fun reportErrorAtPos (_ : Word32.word) {fileName, fileStartLoc = {line, pos}, fileEndLoc = {line = _ : int, pos = _ : int}} (_ : Word32.word) str =
          (errorReportedRef := (!errorReportedRef) + 1;
           print ("#" ^ (Int.toString (!errorReportedRef)) ^
                  " " ^ fileName ^
                  " line " ^ (Int.toString line) ^
                  " pos " ^ (Int.toString pos) ^
                  ":\n" ^ str ^ "\n"))

    fun reportErrorAtPosStrs m pos code strs =
          reportErrorAtPos m pos code (String.concat strs)

    fun reportInternalError (_ : Word32.word) (_ : Word32.word) str =
          (errorReportedRef := (!errorReportedRef) + 1;
           (* We do not print error number on internal errors,
              because internal errors are not supposed to be "normal" -
              so hopefully it signifies to the user that something is off. *)
           print ("Internal error in type-inference: " ^ str ^ "\n");
           raise InternalError)
    
    fun reportInternalErrorStrs m code strs =
          reportInternalError m code (String.concat strs)

    fun reportInternalErrorAtPos (_ : Word32.word) {fileName, fileStartLoc = {line, pos}, fileEndLoc = {line = _ : int, pos = _ : int}} (_ : Word32.word) str =
          (errorReportedRef := (!errorReportedRef) + 1;
           (* We do not print error number on internal errors,
              because internal errors are not supposed to be "normal" -
              so hopefully it signifies to the user that something is off. *)
           print ("Internal error in type-inference in file " ^ fileName ^
                  " line " ^ (Int.toString line) ^
                  " pos " ^ (Int.toString pos) ^
                  ":\n" ^ str ^ "\n");
           raise InternalError)

    fun reportInternalErrorAtPosStrs m pos code strs =
          reportInternalErrorAtPos m pos code (String.concat strs)
  end
