(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure TestKsiTypeInfer =
  struct
    structure M = KsiTypeInfer
    structure ST = M.ST

    exception AcceptMoreThanOnce
    exception AcceptWrongValue
    exception BuiltinError

    (* Test the lexicographical ordering of type variables *)
    fun testLexicogTyvar () =
          M.lexicogTyvar 1 = "'a" andalso
          M.lexicogTyvar 2 = "'b" andalso
          M.lexicogTyvar 25 = "'y" andalso
          M.lexicogTyvar 26 = "'z" andalso
          M.lexicogTyvar 27 = "'aa" andalso
          M.lexicogTyvar 28 = "'ab" andalso
          M.lexicogTyvar 52 = "'az" andalso
          M.lexicogTyvar 53 = "'ba" andalso
          M.lexicogTyvar 54 = "'bb" andalso
          M.lexicogTyvar 55 = "'bc" andalso
          M.lexicogTyvar 702 = "'zz" andalso
          M.lexicogTyvar 703 = "'aaa" andalso
          M.lexicogTyvar 704 = "'aab"

    (* Test the lexicographically ordered production of type variables *)
    fun testLexicogTyvarProducer () =
          let
            val p = M.lexicogTyvarProducer ()
          in
            p () = "'a" andalso
            p () = "'b" andalso
            p () = "'c"
          end

    fun testImplScopedTySpec () =
          let
            fun hasVar env name =
                  case ST.lookup env (ST.symbol name) of
                    NONE =>
                      false
                  | SOME _ =>
                      true

            val ty = KsiAbs.TyFun
                       {from =
                          KsiAbs.TyCon
                            {args =
                               [{name = "'a",
                                 ty = KsiAbs.TyMeta
                                        {name = "'meta#1",
                                         restr = KsiAbs.RstrNone,
                                         bound =
                                           ref (SOME (KsiAbs.TyVar
                                                       {name = "'b",
                                                        restr = KsiAbs.RstrNone,
                                                        ann = ()})),
                                         ann = ()}}],
                             name = "list",
                             tags = ["::", "nil"],
                             ann = ()},
                        to =
                          KsiAbs.TyVar
                            {name = "'c",
                             restr = KsiAbs.RstrNone,
                             ann = ()},
                        ann = ()}
            
            val tyvE = KsiTypeInfer.implScopedTySpec (ST.empty ()) ty
          in
            TestUtil.doTest
              [("not has 'a", fn () => not (hasVar tyvE "'a")),
               ("has 'b", fn () => hasVar tyvE "'b"),
               ("has 'c", fn () => hasVar tyvE "'c")
              ]
          end

    (* Dummy annotation for the abstract syntax *)
    fun annPos pos =
          {pos = {fileName = "<internal test>",
                  fileStartLoc = {line = 0, pos = pos},
                  fileEndLoc = {line = 0, pos = pos}}}

    (* utility functions for building simple
       operator calls in \xi-Calculus abstract syntax. *)
    fun makePresentField exp =
          KsiAbs.ExpOptField
            {present = KsiAbs.ExpCon {tag = "present",
                                      tyScheme = KsiBuiltinStatic.presentTyScheme,
                                      ann = annPos 1001},
             exp = exp, ann = annPos 1002}

    fun makeKsiExpOpExp exp1 opName exp2 pos =
          KsiAbs.ExpApp
            {funExp = KsiAbs.ExpVar
                        {name = opName,
                         ann = annPos (21000 + pos)},
             argExp = KsiAbs.ExpExtRecord
                        {fields = [{lab = "1",
                                    exp = makePresentField exp1,
                                    ann = annPos (22000 + pos)},
                                   {lab = "2",
                                    exp = makePresentField exp2,
                                    ann = annPos (23000 + pos)}
                                  ],
                         ext = KsiAbs.ExpEmptyRecord {ann = annPos 1004},
                         ann = annPos (24000 + pos)},
             ann = annPos (25000 + pos)}

    fun makeKsiVarOpInt varName opName i pos =
          makeKsiExpOpExp
            (KsiAbs.ExpVar
               {name = varName, ann = annPos (26000 + pos)})
            opName
            (KsiAbs.ExpSCon
               {scon = KsiAbs.SCInt i, ann = annPos (27000 + pos)})
            pos

    fun makeKsiVarOpVar varName1 opName varName2 pos =
          makeKsiExpOpExp
            (KsiAbs.ExpVar
               {name = varName1, ann = annPos (28000 + pos)})
            opName
            (KsiAbs.ExpVar
               {name = varName2, ann = annPos (29000 + pos)})
            pos

    fun makeKsiExpConsExp exp1 tagName tyScheme exp2 =
          KsiAbs.ExpApp
            {funExp = KsiAbs.ExpCon
                        {tag = tagName,
                         tyScheme = tyScheme,
                         ann = annPos 1011},
             argExp = KsiAbs.ExpExtRecord
                        {fields = [{lab = "1",
                                    exp = makePresentField exp1,
                                    ann = annPos 1020},
                                   {lab = "2",
                                    exp = makePresentField exp2,
                                    ann = annPos 1021}
                                  ],
                         ext = KsiAbs.ExpEmptyRecord {ann = annPos 1012},
                         ann = annPos 1013},
             ann = annPos 1014}

    fun makeKsiVarConsVar varName1 tagName tyScheme varName2 =
          makeKsiExpConsExp
            (KsiAbs.ExpVar
               {name = varName1, ann = annPos 1015})
            tagName tyScheme
            (KsiAbs.ExpVar
               {name = varName2, ann = annPos 1016})


    fun makeKsiIntConsExp i tagName tyScheme exp =
          makeKsiExpConsExp
            (KsiAbs.ExpSCon
               {scon = KsiAbs.SCInt i, ann = annPos 1017})
            tagName tyScheme exp
            
    (* Value utility function *)
    fun decons value =
          (case value of
             (KsiVals.Constructor
                  {name = "::",
                   value = SOME (KsiVals.Record fields)}) =>
                     (case fields of
                        [{lab = "1", value = value1},
                         {lab = "2", value = value2}] =>
                           (value1, value2)
                      | [{lab = "2", value = value2},
                         {lab = "1", value = value1}] =>
                           (value1, value2)
                      | _ =>
                           (print ("KsiBuiltinTypeError: " ^ "decons (1)");
                            raise BuiltinError)
             )
           | _ =>
               (print ("KsiBuiltinTypeError: " ^ "decons (2)");
                raise BuiltinError)
          )
    
    (* In test1 we test the equivalent of this CeXL-program,
       where we assume the function accept5 to be predeclared.
       accept5 should be called exactly once with the integer argument 5. *)
    (*
       let
         fun testStop cnt =
               if cnt < 5 then
                 continue cnt
               else
                 accept5 cnt

         and continue cnt =
               testStop (cnt + 1)
       in
         testStop 0
       end
    *)
    
    (* This is the \xi-Calculus abstract syntax for the above program: *)
    val test1KsiExp =
          KsiAbs.ExpLetRec
            {binds = [{pat = KsiAbs.PatVar
                               {name = "testStop", ann = annPos 18},
                       exp = KsiAbs.ExpFun
                               {param = "cnt",
                                exp = KsiAbs.ExpCase
                                        {exp = makeKsiVarOpInt
                                                 "cnt" "<" 5 50,
                                         matches =
                                           [{pat = KsiAbs.PatCon
                                                     {tag = "true",
                                                      tyScheme = KsiBuiltinStatic.trueTyScheme,
                                                      ann = annPos 19},
                                             exp = KsiAbs.ExpApp
                                                     {funExp = KsiAbs.ExpVar
                                                                 {name = "continue",
                                                                  ann = annPos 20},
                                                      argExp = KsiAbs.ExpVar
                                                                 {name = "cnt",
                                                                  ann = annPos 21},
                                                      ann = annPos 22},
                                             ann = annPos 20},
                                            {pat = KsiAbs.PatCon
                                                     {tag = "false",
                                                      tyScheme = KsiBuiltinStatic.falseTyScheme,
                                                      ann = annPos 23},
                                             exp = KsiAbs.ExpApp
                                                     {funExp = KsiAbs.ExpVar
                                                                 {name = "accept5",
                                                                  ann = annPos 24},
                                                      argExp = KsiAbs.ExpVar
                                                                 {name = "cnt",
                                                                  ann = annPos 25},
                                                      ann = annPos 26},
                                             ann = annPos 29}
                                           ],
                                         ann = annPos 27},
                                ann = annPos 28},
                       ann = annPos 29},
                      {pat = KsiAbs.PatVar
                               {name = "continue", ann = annPos 30},
                       exp = KsiAbs.ExpFun
                               {param = "cnt",
                                exp = KsiAbs.ExpApp
                                        {funExp = KsiAbs.ExpVar
                                                    {name = "testStop",
                                                     ann = annPos 31},
                                         argExp = makeKsiVarOpInt
                                                    "cnt" "+" 1 51,
                                         ann = annPos 32},
                                ann = annPos 33},
                       ann = annPos 34}],
             exp = KsiAbs.ExpApp
                     {funExp = KsiAbs.ExpVar
                                 {name = "testStop",
                                  ann = annPos 35},
                      argExp = KsiAbs.ExpSCon
                                 {scon = KsiAbs.SCInt 0,
                                  ann = annPos 36},
                      ann = annPos 37},
             ann = annPos 28}

    (* In test2 we test the equivalent of this CeXL-program,
       where we assume the function acceptList321 to be predeclared.
       acceptList321 should be called exactly once with the argument [3, 2, 1]. *)
    (*
       let
         fun rev l acc =
               case l of
                 (x::xs) =>
                   rev xs (x::acc)
               | nil =>
                   acc
       in
         acceptList321 (rev [1, 2, 3] nil)
       end
    *)
    
    (* This is the \xi-Calculus abstract syntax for the above program: *)
    val test2KsiExp =
          KsiAbs.ExpLetRec
            {binds =
               [{pat = KsiAbs.PatVar
                         {name = "rev", ann = annPos 1},
                 exp = KsiAbs.ExpFun
                         {param = "l",
                          exp = KsiAbs.ExpFun
                                  {param = "acc",
                                   exp = KsiAbs.ExpCase
                                           {exp = KsiAbs.ExpVar
                                                    {name = "l", ann = annPos 2},
                                            matches =
                                              [{pat = KsiAbs.PatParamCon
                                                        {tag = "::",
                                                         pat = KsiAbs.PatExtRecord
                                                                 {fields =
                                                                    [{lab = "1",
                                                                      pat = KsiAbs.PatOptField
                                                                              {present =
                                                                                 KsiAbs.PatCon
                                                                                   {tag = "present",
                                                                                    tyScheme = KsiBuiltinStatic.presentTyScheme,
                                                                                    ann = annPos 3},
                                                                               pat = KsiAbs.PatVar
                                                                                       {name = "x",
                                                                                        ann = annPos 4},
                                                                               ann = annPos 5},
                                                                      ann = annPos 30},
                                                                     {lab = "2",
                                                                      pat = KsiAbs.PatOptField
                                                                              {present =
                                                                                 KsiAbs.PatCon
                                                                                   {tag = "present",
                                                                                    tyScheme = KsiBuiltinStatic.presentTyScheme,
                                                                                    ann = annPos 6},
                                                                               pat = KsiAbs.PatVar
                                                                                       {name = "xs",
                                                                                        ann = annPos 7},
                                                                               ann = annPos 8},
                                                                      ann = annPos 31}],
                                                                  ext = KsiAbs.PatEmptyRecord
                                                                          {ann = annPos 9},
                                                                  ann = annPos 10},
                                                         tyScheme = KsiBuiltinStatic.consTyScheme,
                                                         ann = annPos 11},
                                                exp = KsiAbs.ExpApp
                                                        {funExp = KsiAbs.ExpApp
                                                                    {funExp =
                                                                       KsiAbs.ExpVar
                                                                         {name = "rev",
                                                                          ann = annPos 12},
                                                                     argExp =
                                                                       KsiAbs.ExpVar
                                                                         {name = "xs",
                                                                          ann = annPos 13},
                                                                     ann = annPos 14},
                                                         argExp = makeKsiVarConsVar
                                                                    "x" "::" KsiBuiltinStatic.consTyScheme "acc",
                                                         ann = annPos 15},
                                                ann = annPos 32},
                                               {pat = KsiAbs.PatCon
                                                        {tag = "nil",
                                                         tyScheme = KsiBuiltinStatic.nilTyScheme,
                                                         ann = annPos 16},
                                                exp = KsiAbs.ExpVar
                                                        {name = "acc",
                                                         ann = annPos 17},
                                                ann = annPos 33}
                                              ],
                                            ann = annPos 18},
                                   ann = annPos 19},
                          ann = annPos 20},
                 ann = annPos 21}],
             exp = KsiAbs.ExpApp
                     {funExp = KsiAbs.ExpVar
                                 {name = "acceptList321",
                                  ann = annPos 22},
                      argExp = KsiAbs.ExpApp
                                 {funExp = KsiAbs.ExpApp
                                             {funExp =
                                                KsiAbs.ExpVar
                                                  {name = "rev",
                                                   ann = annPos 23},
                                              argExp =
                                                makeKsiIntConsExp
                                                  1 "::" KsiBuiltinStatic.consTyScheme
                                                  (makeKsiIntConsExp
                                                     2 "::" KsiBuiltinStatic.consTyScheme
                                                     (makeKsiIntConsExp
                                                        3 "::" KsiBuiltinStatic.consTyScheme
                                                        (KsiAbs.ExpCon
                                                           {tag = "nil",
                                                            tyScheme = KsiBuiltinStatic.nilTyScheme,
                                                            ann = annPos 24})
                                                     )
                                                  ),
                                              ann = annPos 25},
                                  argExp = KsiAbs.ExpCon
                                             {tag = "nil",
                                              tyScheme = KsiBuiltinStatic.nilTyScheme,
                                              ann = annPos 26},
                                  ann = annPos 27},
                      ann = annPos 28},
             ann = annPos 29}
    
    (* Create the "built-in" environment and the test facilities *)
    val acceptIntTyScheme =
          {params = [],
           ty = KsiAbs.TyFun
                  {from = KsiBuiltinStatic.intTy,
                   to = KsiBuiltinStatic.unitTy,
                   ann = ()},
           ann = ()}

    val intListTy = KsiAbs.TyCon {args = [{name = "'a",
                                           ty = KsiBuiltinStatic.intTy}],
                                  name = "list",
                                  tags = ["nil", "::"],
                                  ann = ()}

    val acceptList321TyScheme =
          {params = [],
           ty = KsiAbs.TyFun
                  {from = intListTy,
                   to = KsiBuiltinStatic.unitTy,
                   ann = ()},
           ann = ()}

    (* Build a symbol table of the type environment of builtin functions *)
    fun bindTyScheme env name tyScheme =
          ST.insert (ST.symbol name)
                    (tyScheme : unit KsiAbs.tyscheme)
                    env
    
    (* Our initial environment must have these "built-in" functions:
       < , + , accept5, acceptList321 *)
    fun makeInitEnv () =
          let
            val initEnv = KsiBuiltinStatic.declareStaVarEnv ()
            val initEnv = bindTyScheme initEnv "accept5" acceptIntTyScheme
            val initEnv = bindTyScheme initEnv "acceptList321" acceptList321TyScheme
          in
            initEnv
          end
    
    fun test1 () =
          (M.reset ();
           M.transProgramExp (makeInitEnv (), ST.empty()) test1KsiExp;
           not (M.errorReported ()))

    fun test2 () =
          (M.reset ();
           M.transProgramExp (makeInitEnv (), ST.empty()) test2KsiExp;
           not (M.errorReported ()))

    fun test () =
          TestUtil.doTest
            [("testLexicogTyvar", testLexicogTyvar),
             ("testLexicogTyvarProducer", testLexicogTyvarProducer),
             ("testImplScopedTySpec", testImplScopedTySpec),
             ("test1", test1),
             ("test2", test2)
            ]
  end
