(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiVals =
  struct
    type Label = string

    datatype t =
      (* In the constructor for functions (Fun), the interpreter
         is "built into" the t -> t function. This should allow us
         to execute both built-in, compiled and interpreted functions! *)
      Fun of t -> t
    | Const of KsiAbs.scon
    | Record of {lab : Label, value : t} list (* Includes tuples *)
    | Field of t option (* This is enough to represent fields *)
    | Constructor of {name : string, value : t option}
      (* Unique exception names are represented by a unit ref. *)
    | Exception of {name : unit ref, value : t option}
    | Ref of t ref

    (* The types used for the CeXL Basis Library *)
    | Array of t Array.array
    | TextIOOutStream of TextIO.outstream
    | TextIOInStream of TextIO.instream
    | BinIOOutStream of BinIO.outstream
    | BinIOInStream of BinIO.instream

    (* We propagate exceptions by raising them in the interpreter *)
    exception Pack of t
  end
