(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure TestUtil (* : TEST_UTIL *) =
  struct
    val output = print

    (* This module has some ugly imperative stuff because we want
       to keep the _interface_ to this module very simple
       - to encourage as much testing as possible :) *)
    val verbose = ref(true)
    val indent = ref nil : string list ref
  
    fun setVerbose value =
          verbose := value

    fun indentedOutput msg =
          if !verbose then
            (List.app output (!indent);
             output msg; output "\n")
          else
            ()

    fun outputBegin msg =
          (indentedOutput msg;
           indent := ("  " :: (!indent)))

    fun outputEnd msg =
          (indent :=
             (case !indent of
                nil =>
                  (output "Error in FTestUtil.outputEnd - ";
                   output "probably unbalanced calls to ";
                   output "outputBegin and outputEnd";
                   nil)
              | (_::rest) =>
                  rest);
           indentedOutput msg)
    
    fun doTest nil =
          true
      | doTest ((name, f)::fList) =
          let
            val _ = outputBegin ("Testing " ^ name ^ ":")

            val error =
                  if f () then
                    NONE
                  else
                    SOME ("Found error in: " ^ name)
                      handle _ =>
                        (SOME ("Found error (exception raised) in: " ^ name))
          in
            case error of
              NONE =>
                (outputEnd (name ^ " OK!");
                 doTest fList)
            | SOME msg =>
                (outputEnd msg;
                 false)
          end
  end
